# Copyright (c) 2016. Librato, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Librato, Inc. nor the names of project contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL LIBRATO, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# Adapted from https://github.com/phobos182/collectd-elasticsearch
# Copyright 2014 Jeremy Carroll
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import json
import re
import urllib2

import collectd

TIMEOUT = 10


Stat = collections.namedtuple('Stat', ('type', 'path', 'metric'))

STATS = [
    Stat("gauge", ("number_of_nodes",), "number_of_nodes"),
    Stat("gauge", ("number_of_data_nodes",), "number_of_data_nodes"),
    Stat("gauge", ("active_primary_shards",), "active_primary_shards"),
    Stat("gauge", ("active_shards",), "active_shards"),
    Stat("gauge", ("relocating_shards",), "relocating_shards"),
    Stat("gauge", ("initializing_shards",), "initializing_shards"),
    Stat("gauge", ("unassigned_shards",), "unassigned_shards"),

    Stat("counter", ("indices", "store", "throttle_time_in_millis"), "indices.store.throttle-time"),

    Stat("gauge", ("indices", "search", "open_contexts"), "indices.search.open-contexts"),

    Stat("counter", ("indices", "fielddata", "evictions"), "indices.cache.field.evictions"),
    Stat("bytes", ("indices", "fielddata", "memory_size_in_bytes"), "indices.cache.field.size"),

    Stat("counter", ("jvm", "gc", "collectors", "young", "collection_time_in_millis"), "jvm.gc.time"),
    Stat("counter", ("jvm", "gc", "collectors", "young", "collection_count"), "jvm.gc.count"),
    Stat("counter", ("jvm", "gc", "collectors", "old", "collection_time_in_millis"), "jvm.gc.old-time"),
    Stat("counter", ("jvm", "gc", "collectors", "old", "collection_count"), "jvm.gc.old-count"),

    Stat("counter", ("indices", "flush", "total"), "indices.flush.total"),
    Stat("counter", ("indices", "flush", "total_time_in_millis"), "indices.flush.time"),

    Stat("gauge", ("indices", "merges", "current"), "indices.merges.current-docs"),
    Stat("bytes", ("indices", "merges", "current_size_in_bytes"), "indices.merges.current-size"),
    Stat("counter", ("indices", "merges", "total"), "indices.merges.total"),
    Stat("gauge", ("indices", "merges", "total_docs"), "indices.merges.total-docs"),
    Stat("bytes", ("indices", "merges", "total_size_in_bytes"), "indices.merges.total-size"),
    Stat("counter", ("indices", "merges", "total_time_in_millis"), "indices.merges.time"),

    Stat("gauge", ("indices", "segments", "count"), "indices.segments.count"),

    Stat("counter", ("indices", "refresh", "total"), "indices.refresh.total"),
    Stat("counter", ("indices", "refresh", "total_time_in_millis"), "indices.refresh.time"),

    Stat("gauge", ("process", "cpu", "percent"), "process.cpu.percent"),

    Stat("gauge", ("indices", "docs", "count"), "indices.docs.count"),
    Stat("counter", ("indices", "docs", "deleted"), "indices.docs.deleted"),

    Stat("bytes", ("indices", "store", "size_in_bytes"), "indices.store.size"),

    Stat("counter", ("indices", "indexing", "index_total"), "indices.indexing.index-total"),
    Stat("counter", ("indices", "indexing", "index_time_in_millis"), "indices.indexing.index-time"),
    Stat("counter", ("indices", "indexing", "delete_total"), "indices.indexing.delete-total"),
    Stat("counter", ("indices", "indexing", "delete_time_in_millis"), "indices.indexing.delete-time"),
    Stat("gauge", ("indices", "indexing", "index_current"), "indices.indexing.index-current"),
    Stat("gauge", ("indices", "indexing", "delete_current"), "indices.indexing.delete-current"),

    Stat("counter", ("indices", "get", "total"), "indices.get.total"),
    Stat("counter", ("indices", "get", "time_in_millis"), "indices.get.time"),
    Stat("counter", ("indices", "get", "exists_total"), "indices.get.exists-total"),
    Stat("counter", ("indices", "get", "exists_time_in_millis"), "indices.get.exists-time"),
    Stat("counter", ("indices", "get", "missing_total"), "indices.get.missing-total"),
    Stat("counter", ("indices", "get", "missing_time_in_millis"), "indices.get.missing-time"),
    Stat("gauge", ("indices", "get", "current"), "indices.get.current"),

    Stat("gauge", ("indices", "search", "query_current"), "indices.search.query-current"),
    Stat("counter", ("indices", "search", "query_total"), "indices.search.query-total"),
    Stat("counter", ("indices", "search", "query_time_in_millis"), "indices.search.query-time"),
    Stat("gauge", ("indices", "search", "fetch_current"), "indices.search.fetch-current"),
    Stat("counter", ("indices", "search", "fetch_total"), "indices.search.fetch-total"),
    Stat("counter", ("indices", "search", "fetch_time_in_millis"), "indices.search.fetch-time"),

    Stat("bytes", ("jvm", "mem", "heap_committed_in_bytes"), "jvm.mem.heap-committed"),
    Stat("bytes", ("jvm", "mem", "heap_used_in_bytes"), "jvm.mem.heap-used"),
    Stat("percent", ("jvm", "mem", "heap_used_percent"), "jvm.mem.heap-used-percent"),
    Stat("bytes", ("jvm", "mem", "non_heap_committed_in_bytes"), "jvm.mem.non-heap-committed"),
    Stat("bytes", ("jvm", "mem", "non_heap_used_in_bytes"), "jvm.mem.non-heap-used"),

    Stat("gauge", ("jvm", "threads", "count"), "jvm.threads.count"),
    Stat("gauge", ("jvm", "threads", "peak_count"), "jvm.threads.peak"),

    Stat("gauge", ("transport", "server_open"), "transport.server_open"),
    Stat("counter", ("transport", "rx_count"), "transport.rx.count"),
    Stat("counter", ("transport", "rx_size_in_bytes"), "transport.rx.bytes"),
    Stat("counter", ("transport", "tx_count"), "transport.tx.count"),
    Stat("counter", ("transport", "tx_size_in_bytes"), "transport.tx.bytes"),

    Stat("gauge", ("http", "current_open"), "http.current_open"),
    Stat("counter", ("http", "total_opened"), "http.total_open"),

    Stat("gauge", ("process", "open_file_descriptors"), "process.open_file_descriptors"),

]


class ElasticsearchPlugin:
    def __init__(self):
        self.version = None
        self.name = None
        self.url = "http://localhost:9200"
        self.verbose = False
        self.cluster_health_url = None
        self.node_stats_url = None

    def log_verbose(self, msg):
        if self.verbose:
            collectd.info("Elasticsearch plugin: %s" % msg)

    def configure_callback(self, conf):
        for node in conf.children:
            if node.key == 'Url':
                self.url = node.values[0]
            elif node.key == 'Name':
                self.name = node.values[0]
            elif node.key == 'Verbose':
                self.verbose = bool(node.values[0])
            else:
                collectd.warning('Elasticsearch plugin: Unknown config key: %s.' % node.key)

        self.log_verbose("url = %s" % self.url)
        if self.name:
            self.log_verbose("name = %s" % self.name)

    def compute_urls(self):
        if not self.version:
            base_url = self.url.rstrip('/')
            doc = json.load(urllib2.urlopen(base_url, timeout=10))
            self.version = doc["version"]["number"]
            self.log_verbose("ES version = %s" % self.version)

            if not self.name:
                cluster_name = doc.get("cluster_name", "elasticsearch")
                self.name = re.sub(r"\s+", '-', cluster_name)
                self.log_verbose("name = %s" % self.name)

            self.cluster_health_url = base_url + '/_cluster/health'
            if self.version.startswith("0."):
                # ElasticSearch 0.9.x
                self.node_stats_url = base_url + "/_cluster/nodes/_local/stats?" \
                "http=true&process=true&jvm=true&transport=true&thread_pool=true"

                STATS.append(Stat("counter", ("jvm", "gc", "collection_time_in_millis"), "jvm.gc.time"))
                STATS.append(Stat("counter", ("jvm", "gc", "collection_count"), "jvm.gc.count"))
            else:
                self.node_stats_url = base_url + "/_nodes/_local/stats/" \
                "transport,http,process,jvm,indices,thread_pool"

                if self.version.startswith("1."):
                    STATS.append(Stat("counter", ("indices", "filter_cache", "evictions"),
                                      "indices.cache.filter.evictions"))
                    STATS.append(Stat("bytes", ("indices", "filter_cache", "memory_size_in_bytes"),
                                      "indices.cache.filter.size"))
                else:
                    STATS.append(Stat("counter", ("indices", "query_cache", "evictions"),
                                      "indices.cache.query.evictions"))
                    STATS.append(Stat("bytes", ("indices", "query_cache", "memory_size_in_bytes"),
                                      "indices.cache.query.size"))
                    STATS.append(Stat("counter", ("indices", "query_cache", "hit_count"),
                                      "indices.cache.query.hits"))
                    STATS.append(Stat("counter", ("indices", "query_cache", "miss_count"),
                                      "indices.cache.query.misses"))

    def fetch_stats(self):
        stats = json.load(urllib2.urlopen(self.cluster_health_url, timeout=10))
        doc = json.load(urllib2.urlopen(self.node_stats_url, timeout=10))
        stats.update(doc["nodes"].values()[0])

        return stats

    def extract_stat(self, path, stats):
        try:
            return int(reduce(lambda x, y: x[y], path, stats))
        except Exception as e:
            collectd.warning("Elasticsearch plugin: Value not found for %s: (%s) %s" %
                             ('.'.join(path), e.__class__.__name__, e))
            return None

    def dispatch_stat(self, type_, instance, value_):
        try:
            self.log_verbose('Sending value[%s]: %s=%s' % (type_, instance, value_))
            val = collectd.Values(plugin='elasticsearch')
            val.plugin_instance = self.name
            val.type = type_
            val.type_instance = instance
            val.values = [value_]
            val.dispatch()
        except Exception as e:
            collectd.warning("Elasticsearch plugin: Error dispatching metric %s: (%s) %s" %
                             instance, e.__class__.__name__, e)

    def report_stats(self, stats):
        if 'status' in stats:
            type_instance = 'status.' + stats['status']	 # Green, yellow or red
            self.dispatch_stat("gauge", type_instance, 1)

        for stat in STATS:
            result = self.extract_stat(stat.path, stats)
            if result is not None:
                self.dispatch_stat(stat.type, stat.metric, result)

        # thread pool metrics vary depending on Elasticsearch version
        if "thread_pool" in stats:
            for pool in stats["thread_pool"]:
                for attr in ['threads', 'queue', 'active', 'largest']:
                    path = ("thread_pool", pool, attr)
                    result = self.extract_stat(path, stats)
                    if result is not None:
                        self.dispatch_stat("gauge", '.'.join(path), result)

                for attr in ['completed', 'rejected']:
                    path = ("thread_pool", pool, attr)
                    result = self.extract_stat(path, stats)
                    if result is not None:
                        self.dispatch_stat("counter", '.'.join(path), result)

    def read_callback(self):
        try:
            self.compute_urls()
            stats = self.fetch_stats()
            self.report_stats(stats)
        except urllib2.URLError as err:
            collectd.error("Elasticsearch plugin: read-function failed: %s" % err.reason)
            return


plugin = ElasticsearchPlugin()


def configure_callback(conf):
    plugin.configure_callback(conf)


def read_callback():
    plugin.read_callback()


collectd.register_config(configure_callback)
collectd.register_read(read_callback)

