# Copyright (c) Sebastien Estienne
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# See https://github.com/sebest/collectd-mongodb

# Copyright (c) 2016. Librato, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Librato, Inc. nor the names of project contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL LIBRATO, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Librato Modifications
# - Rename metrics so they remain meaningful when the plugin instance is stripped away by the backend
#   (a number of them are of type counter and gauge)
# - Prefix metrics with 'mongodb' instead of 'mongo'
# - Change default plugin instance to 'mongodb' instead of the port and provide a 'Name' param to overrride it
# - Add uptime, asserts, globalLock.currentQueue and replication lag metrics
# - Fix lock ratio to work on MongoDB 3.x

#
# Plugin to collectd statistics from MongoDB
#

from datetime import datetime
from collections import defaultdict
import collectd
from pymongo import MongoClient
from pymongo.read_preferences import ReadPreference
from distutils.version import StrictVersion as V


class MongoDB(object):

    def __init__(self):
        self.plugin_name = "mongodb"
        self.plugin_instance = "mongodb"
        self.mongo_host = "127.0.0.1"
        self.mongo_port = 27017
        self.mongo_db = ["admin", ]
        self.mongo_user = None
        self.mongo_password = None

        self.lockTotalTime = None
        self.lockTime = None
        self.accesses = None
        self.misses = None

    def submit(self, type, instance, value, db=None):
        if db:
            plugin_instance = '%s-%s' % (self.plugin_instance, db)
        else:
            plugin_instance = self.plugin_instance
        v = collectd.Values()
        v.plugin = self.plugin_name
        v.plugin_instance = plugin_instance
        v.type = type
        v.type_instance = instance
        v.values = [value, ]
        v.dispatch()

    def do_server_status(self, con, db):
        server_status = db.command('serverStatus')

        version = server_status['version']
        at_least_2_4 = V(version) >= V('2.4.0')

        # uptime
        if 'uptimeMillis' in server_status:
            self.submit('gauge', 'uptime_millis', server_status['uptimeMillis'])

        # extra info
        if 'page_faults' in server_status['extra_info']:
            self.submit('counter', 'page_faults', server_status['extra_info']['page_faults'])

        # operations
        for k, v in server_status['opcounters'].items():
            self.submit('total_operations', 'total_operations.' + k, v)

        # memory
        for t in ['resident', 'virtual', 'mapped']:
            self.submit('memory', 'mem.' + t, server_status['mem'][t])

        # connections
        self.submit('connections', 'connections.current', server_status['connections']['current'])
        if 'available' in server_status['connections']:
            self.submit('connections', 'connections.available', server_status['connections']['available'])
        if 'totalCreated' in server_status['connections']:
            self.submit('connections', 'connections.totalCreated', server_status['connections']['totalCreated'])

        # network
        if 'network' in server_status:
            for t in ['bytesIn', 'bytesOut', 'numRequests']:
                self.submit('counter', 'network.' + t, server_status['network'][t])

        # locks
        lock_time = None
        if 'lockTime' in server_status['globalLock']:
            lock_time = server_status['globalLock']['lockTime']
        elif 'locks' in server_status and 'Global' in server_status['locks']:
            if 'timeAcquiringMicros' in server_status['locks']['Global']:
                lock_time = 0
                for type_ in server_status['locks']['Global']['timeAcquiringMicros']:
                    lock_time += server_status['locks']['Global']['timeAcquiringMicros'][type_]

        if lock_time is not None:
            if self.lockTotalTime is not None and self.lockTime is not None:
                if self.lockTime == lock_time:
                    value = 0.0
                else:
                    value = float(lock_time - self.lockTime) * 100.0 / float(server_status['globalLock']['totalTime'] - self.lockTotalTime)
                self.submit('percent', 'lock_ratio', value)

            self.lockTime = lock_time
        self.lockTotalTime = server_status['globalLock']['totalTime']

        # currentQueue
        if 'currentQueue' in server_status['globalLock']:
            for t in ['readers', 'writers', 'total']:
                self.submit('gauge', 'lock.current_queue.' + t, server_status['globalLock']['currentQueue'][t])

        # indexes
        if 'indexCounters' in server_status:
            accesses = None
            misses = None
            index_counters = server_status['indexCounters'] if at_least_2_4 else server_status['indexCounters']['btree']

            if self.accesses is not None:
                accesses = index_counters['accesses'] - self.accesses
                if accesses < 0:
                    accesses = None
            misses = (index_counters['misses'] or 0) - (self.misses or 0)
            if misses < 0:
                misses = None
            if accesses and misses is not None:
                self.submit('cache_ratio', 'cache_ratio.cache_misses', int(misses * 100 / float(accesses)))
            else:
                self.submit('cache_ratio', 'cache_ratio.cache_misses', 0)
            self.accesses = index_counters['accesses']
            self.misses = index_counters['misses']

        # asserts
        if 'asserts' in server_status:
            for t in server_status['asserts']:
                self.submit('counter', 'asserts.' + t, server_status['asserts'][t])

        for mongo_db in self.mongo_db:
            db = con[mongo_db]
            if self.mongo_user and self.mongo_password:
                con[self.mongo_db[0]].authenticate(self.mongo_user, self.mongo_password)
            db_stats = db.command('dbstats')

            # stats counts
            self.submit('counter', 'object_count', db_stats['objects'], mongo_db)
            self.submit('counter', 'collections', db_stats['collections'], mongo_db)
            self.submit('counter', 'num_extents', db_stats['numExtents'], mongo_db)
            self.submit('counter', 'indexes', db_stats['indexes'], mongo_db)

            # stats sizes
            self.submit('file_size', 'storage_size', db_stats['storageSize'], mongo_db)
            self.submit('file_size', 'index_size', db_stats['indexSize'], mongo_db)
            self.submit('file_size', 'data_size', db_stats['dataSize'], mongo_db)

    def do_repl_status(self, db):
        try:
            repl_status = db.command('replSetGetStatus')
        except Exception as e:
            collectd.info("mongodb plugin: Error in replSetGetStatus: %s" % e)
            return

        secondary_optimes = []
        primary_optime = None
        node_states = defaultdict(int)

        for member in repl_status['members']:
            node_states[member['stateStr']] += 1
            if member.get('state') == 1 and member.get('self') == True:
                # Only report the lag if the current node is primary
                primary_optime = member.get('optimeDate')
            elif member.get('state') != 7:
                # Skip arbiters
                sec_optime = member.get('optimeDate')
                secondary_optimes.append(sec_optime)

        if primary_optime is not None:
            lag = sum([(primary_optime - s).total_seconds() for s in secondary_optimes])
            self.submit('gauge', 'replication.lag_millis', lag*1000)

        for st in node_states:
            self.submit('gauge', 'node_state.' + st.lower(), node_states[st])

    def read(self):
        con = None
        try:
            con = MongoClient(host=self.mongo_host, port=self.mongo_port,
                              read_preference=ReadPreference.SECONDARY)
            db = con[self.mongo_db[0]]
            if self.mongo_user and self.mongo_password:
                db.authenticate(self.mongo_user, self.mongo_password)

            self.do_server_status(con, db)
            self.do_repl_status(db)
        except Exception as e:
            collectd.error("mongodb plugin: read callback failed: %s" % e)
        finally:
            if con:
                con.close()

    def config(self, obj):
        for node in obj.children:
            if node.key == 'Port':
                self.mongo_port = int(node.values[0])
            elif node.key == 'Host':
                self.mongo_host = node.values[0]
            elif node.key == 'User':
                self.mongo_user = node.values[0]
            elif node.key == 'Password':
                self.mongo_password = node.values[0]
            elif node.key == 'Database':
                self.mongo_db = node.values
            elif node.key == 'Name':
                self.plugin_instance = node.values[0]
            else:
                collectd.warning("mongodb plugin: Unknown configuration key %s" % node.key)

mongodb = MongoDB()
collectd.register_read(mongodb.read)
collectd.register_config(mongodb.config)
