import collectd
import urllib2
import time
import sys
import json
import collections

CONFIG = {'Verbose': False}
Measure = collections.namedtuple('Measure', ('name', 'source', 'value'))

def read_callback():
    try:
        get_metrics(CONFIG['URL'])
    except urllib2.URLError as err:
        collectd.error('nginx-plus plugin: read-function failed: %s' % err.reason)
        return

def get_metrics(url):
    try:
        response = urllib2.urlopen(url)
    except urllib2.HTTPError as e:
        collectd.error('nginx-plus plugin: Fetching stats failed: %s %s' % (e.code, e.reason))
        return

    stats = json.loads(response.read())

    submit_value(Measure('generation', None, stats['generation']))

    # Simple sections
    for section in ['processes', 'connections', 'requests', 'ssl']:
        for key, value in stats[section].iteritems():
            collectd.warning('%s' % type(section))
            collectd.warning('%s' % type(key))
            submit_value(Measure(join(section, key), None, value))

    # These sections contain metrics per-zone/cache/upstream
    for section in ['caches', 'server_zones']:
        # config_name is the name of a particular cache config or server_zone
        for config_name, metrics in stats[section].iteritems():
            for key, metrics in metrics.iteritems():
                if type(metrics) == dict: # Go deeper
                    for subkey, value in metrics.iteritems():
                        submit_value(Measure(join(section, key, subkey), config_name, value))
                else:
                    submit_value(Measure(join(section, key), config_name, metrics))

    # Each upstream configuration contains one or more peers (servers) containing metrics
    for config_name, upstream in stats['upstreams'].iteritems():
        for peer in upstream['peers']:
            plugin_instance = config_name + '.' + peer.pop('server')
            for key, metrics in peer.iteritems():
                if type(metrics) == dict: # Go deeper
                    for subkey, value in metrics.iteritems():
                        submit_value(Measure(join('upstreams', key, subkey), plugin_instance, value))
                else:
                    submit_value(Measure(join('upstreams', key), plugin_instance, metrics))

    if 'stream' in stats:
        for config_name, zone in stats['stream']['server_zones'].iteritems():
            for key, value in zone.iteritems():
                submit_value(Measure(join('stream.server_zones', key), config_name, value))

        for config_name, upstream in stats['stream']['upstreams'].iteritems():
            for peer in upstream['peers']:
                plugin_instance = config_name + '.' + peer.pop('server')
                for key, value in peer.iteritems():
                    if type(value) == dict:
                        for subkey, value in value.iteritems():
                            submit_value(Measure(join('stream.upstreams', key, subkey), plugin_instance, value))
                    else:
                        submit_value(Measure(join('stream.upstreams', key), plugin_instance, value))

def configure_callback(conf):
    for node in conf.children:
        if node.key == 'URL':
            CONFIG['URL'] = node.values[0]
            collectd.notice('nginx-plus plugin: Setting nginx-plus stats URL to %s' % node.values[0] )
        elif node.key == 'Verbose':
            CONFIG['Verbose'] = node.values[0]
        else:
            log('nginx-plus plugin: Unknown config key: %s.' % node.key )

# Only float or integers should be allowed
def reject_invalid_datatype(value):
    return isinstance(value, (bool,str,unicode))

# Some Nginx Plus metrics don't make map to gauges well
def reject_metric_blacklist(metric_name):
    blacklist = ['weight','downstart', 'selected', 'id', 'max_size', 'max_conns']
    parts = metric_name.split('.')
    for part in parts:
        for term in blacklist:
            if term == part:
                return True
    return False

def detect_type(name):
    # These metrics are point-in-time snapshots. All others are running totals
    gauges = ['processing', 'active', 'idle', 'current', 'size', 'keepalive',
            'connect_time', 'first_byte_time', 'response_time']
    for gauge in gauges:
        if gauge in name:
            return 'gauge'
    return 'derive'

def submit_value(measure):
    if reject_invalid_datatype(measure.value):
        log("Can't submit value '%s' of type '%s'" % (measure.value, type(measure.value)))
        return

    if reject_metric_blacklist(measure.name):
        log("Skipping blacklisted metric '%s'" % measure.name)
        return

    v = collectd.Values(plugin='nginx_plus')
    v.type = detect_type(measure.name)
    v.type_instance = measure.name
    if measure.source:
        v.plugin_instance = measure.source
    v.values = [measure.value]
    log('%s' % v)
    v.dispatch()

def log(message):
    if CONFIG['Verbose']:
        collectd.warning(message)

def join(*parts):
    return '.'.join(parts)

collectd.register_config(configure_callback)
collectd.register_read(read_callback)

